Ian Goldberg 14 роки тому
коміт
e2973619c9
9 змінених файлів з 2247 додано та 0 видалено
  1. 674 0
      COPYING
  2. 50 0
      Makefile
  3. 46 0
      README
  4. 31 0
      cudadl.h
  5. 256 0
      dlrho.cc
  6. 121 0
      dpstream.cu
  7. 145 0
      gen_N.cc
  8. 323 0
      gencios_reg_20
  9. 601 0
      parrhoasm.cu

+ 674 - 0
COPYING

@@ -0,0 +1,674 @@
+                    GNU GENERAL PUBLIC LICENSE
+                       Version 3, 29 June 2007
+
+ Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
+ Everyone is permitted to copy and distribute verbatim copies
+ of this license document, but changing it is not allowed.
+
+                            Preamble
+
+  The GNU General Public License is a free, copyleft license for
+software and other kinds of works.
+
+  The licenses for most software and other practical works are designed
+to take away your freedom to share and change the works.  By contrast,
+the GNU General Public License is intended to guarantee your freedom to
+share and change all versions of a program--to make sure it remains free
+software for all its users.  We, the Free Software Foundation, use the
+GNU General Public License for most of our software; it applies also to
+any other work released this way by its authors.  You can apply it to
+your programs, too.
+
+  When we speak of free software, we are referring to freedom, not
+price.  Our General Public Licenses are designed to make sure that you
+have the freedom to distribute copies of free software (and charge for
+them if you wish), that you receive source code or can get it if you
+want it, that you can change the software or use pieces of it in new
+free programs, and that you know you can do these things.
+
+  To protect your rights, we need to prevent others from denying you
+these rights or asking you to surrender the rights.  Therefore, you have
+certain responsibilities if you distribute copies of the software, or if
+you modify it: responsibilities to respect the freedom of others.
+
+  For example, if you distribute copies of such a program, whether
+gratis or for a fee, you must pass on to the recipients the same
+freedoms that you received.  You must make sure that they, too, receive
+or can get the source code.  And you must show them these terms so they
+know their rights.
+
+  Developers that use the GNU GPL protect your rights with two steps:
+(1) assert copyright on the software, and (2) offer you this License
+giving you legal permission to copy, distribute and/or modify it.
+
+  For the developers' and authors' protection, the GPL clearly explains
+that there is no warranty for this free software.  For both users' and
+authors' sake, the GPL requires that modified versions be marked as
+changed, so that their problems will not be attributed erroneously to
+authors of previous versions.
+
+  Some devices are designed to deny users access to install or run
+modified versions of the software inside them, although the manufacturer
+can do so.  This is fundamentally incompatible with the aim of
+protecting users' freedom to change the software.  The systematic
+pattern of such abuse occurs in the area of products for individuals to
+use, which is precisely where it is most unacceptable.  Therefore, we
+have designed this version of the GPL to prohibit the practice for those
+products.  If such problems arise substantially in other domains, we
+stand ready to extend this provision to those domains in future versions
+of the GPL, as needed to protect the freedom of users.
+
+  Finally, every program is threatened constantly by software patents.
+States should not allow patents to restrict development and use of
+software on general-purpose computers, but in those that do, we wish to
+avoid the special danger that patents applied to a free program could
+make it effectively proprietary.  To prevent this, the GPL assures that
+patents cannot be used to render the program non-free.
+
+  The precise terms and conditions for copying, distribution and
+modification follow.
+
+                       TERMS AND CONDITIONS
+
+  0. Definitions.
+
+  "This License" refers to version 3 of the GNU General Public License.
+
+  "Copyright" also means copyright-like laws that apply to other kinds of
+works, such as semiconductor masks.
+
+  "The Program" refers to any copyrightable work licensed under this
+License.  Each licensee is addressed as "you".  "Licensees" and
+"recipients" may be individuals or organizations.
+
+  To "modify" a work means to copy from or adapt all or part of the work
+in a fashion requiring copyright permission, other than the making of an
+exact copy.  The resulting work is called a "modified version" of the
+earlier work or a work "based on" the earlier work.
+
+  A "covered work" means either the unmodified Program or a work based
+on the Program.
+
+  To "propagate" a work means to do anything with it that, without
+permission, would make you directly or secondarily liable for
+infringement under applicable copyright law, except executing it on a
+computer or modifying a private copy.  Propagation includes copying,
+distribution (with or without modification), making available to the
+public, and in some countries other activities as well.
+
+  To "convey" a work means any kind of propagation that enables other
+parties to make or receive copies.  Mere interaction with a user through
+a computer network, with no transfer of a copy, is not conveying.
+
+  An interactive user interface displays "Appropriate Legal Notices"
+to the extent that it includes a convenient and prominently visible
+feature that (1) displays an appropriate copyright notice, and (2)
+tells the user that there is no warranty for the work (except to the
+extent that warranties are provided), that licensees may convey the
+work under this License, and how to view a copy of this License.  If
+the interface presents a list of user commands or options, such as a
+menu, a prominent item in the list meets this criterion.
+
+  1. Source Code.
+
+  The "source code" for a work means the preferred form of the work
+for making modifications to it.  "Object code" means any non-source
+form of a work.
+
+  A "Standard Interface" means an interface that either is an official
+standard defined by a recognized standards body, or, in the case of
+interfaces specified for a particular programming language, one that
+is widely used among developers working in that language.
+
+  The "System Libraries" of an executable work include anything, other
+than the work as a whole, that (a) is included in the normal form of
+packaging a Major Component, but which is not part of that Major
+Component, and (b) serves only to enable use of the work with that
+Major Component, or to implement a Standard Interface for which an
+implementation is available to the public in source code form.  A
+"Major Component", in this context, means a major essential component
+(kernel, window system, and so on) of the specific operating system
+(if any) on which the executable work runs, or a compiler used to
+produce the work, or an object code interpreter used to run it.
+
+  The "Corresponding Source" for a work in object code form means all
+the source code needed to generate, install, and (for an executable
+work) run the object code and to modify the work, including scripts to
+control those activities.  However, it does not include the work's
+System Libraries, or general-purpose tools or generally available free
+programs which are used unmodified in performing those activities but
+which are not part of the work.  For example, Corresponding Source
+includes interface definition files associated with source files for
+the work, and the source code for shared libraries and dynamically
+linked subprograms that the work is specifically designed to require,
+such as by intimate data communication or control flow between those
+subprograms and other parts of the work.
+
+  The Corresponding Source need not include anything that users
+can regenerate automatically from other parts of the Corresponding
+Source.
+
+  The Corresponding Source for a work in source code form is that
+same work.
+
+  2. Basic Permissions.
+
+  All rights granted under this License are granted for the term of
+copyright on the Program, and are irrevocable provided the stated
+conditions are met.  This License explicitly affirms your unlimited
+permission to run the unmodified Program.  The output from running a
+covered work is covered by this License only if the output, given its
+content, constitutes a covered work.  This License acknowledges your
+rights of fair use or other equivalent, as provided by copyright law.
+
+  You may make, run and propagate covered works that you do not
+convey, without conditions so long as your license otherwise remains
+in force.  You may convey covered works to others for the sole purpose
+of having them make modifications exclusively for you, or provide you
+with facilities for running those works, provided that you comply with
+the terms of this License in conveying all material for which you do
+not control copyright.  Those thus making or running the covered works
+for you must do so exclusively on your behalf, under your direction
+and control, on terms that prohibit them from making any copies of
+your copyrighted material outside their relationship with you.
+
+  Conveying under any other circumstances is permitted solely under
+the conditions stated below.  Sublicensing is not allowed; section 10
+makes it unnecessary.
+
+  3. Protecting Users' Legal Rights From Anti-Circumvention Law.
+
+  No covered work shall be deemed part of an effective technological
+measure under any applicable law fulfilling obligations under article
+11 of the WIPO copyright treaty adopted on 20 December 1996, or
+similar laws prohibiting or restricting circumvention of such
+measures.
+
+  When you convey a covered work, you waive any legal power to forbid
+circumvention of technological measures to the extent such circumvention
+is effected by exercising rights under this License with respect to
+the covered work, and you disclaim any intention to limit operation or
+modification of the work as a means of enforcing, against the work's
+users, your or third parties' legal rights to forbid circumvention of
+technological measures.
+
+  4. Conveying Verbatim Copies.
+
+  You may convey verbatim copies of the Program's source code as you
+receive it, in any medium, provided that you conspicuously and
+appropriately publish on each copy an appropriate copyright notice;
+keep intact all notices stating that this License and any
+non-permissive terms added in accord with section 7 apply to the code;
+keep intact all notices of the absence of any warranty; and give all
+recipients a copy of this License along with the Program.
+
+  You may charge any price or no price for each copy that you convey,
+and you may offer support or warranty protection for a fee.
+
+  5. Conveying Modified Source Versions.
+
+  You may convey a work based on the Program, or the modifications to
+produce it from the Program, in the form of source code under the
+terms of section 4, provided that you also meet all of these conditions:
+
+    a) The work must carry prominent notices stating that you modified
+    it, and giving a relevant date.
+
+    b) The work must carry prominent notices stating that it is
+    released under this License and any conditions added under section
+    7.  This requirement modifies the requirement in section 4 to
+    "keep intact all notices".
+
+    c) You must license the entire work, as a whole, under this
+    License to anyone who comes into possession of a copy.  This
+    License will therefore apply, along with any applicable section 7
+    additional terms, to the whole of the work, and all its parts,
+    regardless of how they are packaged.  This License gives no
+    permission to license the work in any other way, but it does not
+    invalidate such permission if you have separately received it.
+
+    d) If the work has interactive user interfaces, each must display
+    Appropriate Legal Notices; however, if the Program has interactive
+    interfaces that do not display Appropriate Legal Notices, your
+    work need not make them do so.
+
+  A compilation of a covered work with other separate and independent
+works, which are not by their nature extensions of the covered work,
+and which are not combined with it such as to form a larger program,
+in or on a volume of a storage or distribution medium, is called an
+"aggregate" if the compilation and its resulting copyright are not
+used to limit the access or legal rights of the compilation's users
+beyond what the individual works permit.  Inclusion of a covered work
+in an aggregate does not cause this License to apply to the other
+parts of the aggregate.
+
+  6. Conveying Non-Source Forms.
+
+  You may convey a covered work in object code form under the terms
+of sections 4 and 5, provided that you also convey the
+machine-readable Corresponding Source under the terms of this License,
+in one of these ways:
+
+    a) Convey the object code in, or embodied in, a physical product
+    (including a physical distribution medium), accompanied by the
+    Corresponding Source fixed on a durable physical medium
+    customarily used for software interchange.
+
+    b) Convey the object code in, or embodied in, a physical product
+    (including a physical distribution medium), accompanied by a
+    written offer, valid for at least three years and valid for as
+    long as you offer spare parts or customer support for that product
+    model, to give anyone who possesses the object code either (1) a
+    copy of the Corresponding Source for all the software in the
+    product that is covered by this License, on a durable physical
+    medium customarily used for software interchange, for a price no
+    more than your reasonable cost of physically performing this
+    conveying of source, or (2) access to copy the
+    Corresponding Source from a network server at no charge.
+
+    c) Convey individual copies of the object code with a copy of the
+    written offer to provide the Corresponding Source.  This
+    alternative is allowed only occasionally and noncommercially, and
+    only if you received the object code with such an offer, in accord
+    with subsection 6b.
+
+    d) Convey the object code by offering access from a designated
+    place (gratis or for a charge), and offer equivalent access to the
+    Corresponding Source in the same way through the same place at no
+    further charge.  You need not require recipients to copy the
+    Corresponding Source along with the object code.  If the place to
+    copy the object code is a network server, the Corresponding Source
+    may be on a different server (operated by you or a third party)
+    that supports equivalent copying facilities, provided you maintain
+    clear directions next to the object code saying where to find the
+    Corresponding Source.  Regardless of what server hosts the
+    Corresponding Source, you remain obligated to ensure that it is
+    available for as long as needed to satisfy these requirements.
+
+    e) Convey the object code using peer-to-peer transmission, provided
+    you inform other peers where the object code and Corresponding
+    Source of the work are being offered to the general public at no
+    charge under subsection 6d.
+
+  A separable portion of the object code, whose source code is excluded
+from the Corresponding Source as a System Library, need not be
+included in conveying the object code work.
+
+  A "User Product" is either (1) a "consumer product", which means any
+tangible personal property which is normally used for personal, family,
+or household purposes, or (2) anything designed or sold for incorporation
+into a dwelling.  In determining whether a product is a consumer product,
+doubtful cases shall be resolved in favor of coverage.  For a particular
+product received by a particular user, "normally used" refers to a
+typical or common use of that class of product, regardless of the status
+of the particular user or of the way in which the particular user
+actually uses, or expects or is expected to use, the product.  A product
+is a consumer product regardless of whether the product has substantial
+commercial, industrial or non-consumer uses, unless such uses represent
+the only significant mode of use of the product.
+
+  "Installation Information" for a User Product means any methods,
+procedures, authorization keys, or other information required to install
+and execute modified versions of a covered work in that User Product from
+a modified version of its Corresponding Source.  The information must
+suffice to ensure that the continued functioning of the modified object
+code is in no case prevented or interfered with solely because
+modification has been made.
+
+  If you convey an object code work under this section in, or with, or
+specifically for use in, a User Product, and the conveying occurs as
+part of a transaction in which the right of possession and use of the
+User Product is transferred to the recipient in perpetuity or for a
+fixed term (regardless of how the transaction is characterized), the
+Corresponding Source conveyed under this section must be accompanied
+by the Installation Information.  But this requirement does not apply
+if neither you nor any third party retains the ability to install
+modified object code on the User Product (for example, the work has
+been installed in ROM).
+
+  The requirement to provide Installation Information does not include a
+requirement to continue to provide support service, warranty, or updates
+for a work that has been modified or installed by the recipient, or for
+the User Product in which it has been modified or installed.  Access to a
+network may be denied when the modification itself materially and
+adversely affects the operation of the network or violates the rules and
+protocols for communication across the network.
+
+  Corresponding Source conveyed, and Installation Information provided,
+in accord with this section must be in a format that is publicly
+documented (and with an implementation available to the public in
+source code form), and must require no special password or key for
+unpacking, reading or copying.
+
+  7. Additional Terms.
+
+  "Additional permissions" are terms that supplement the terms of this
+License by making exceptions from one or more of its conditions.
+Additional permissions that are applicable to the entire Program shall
+be treated as though they were included in this License, to the extent
+that they are valid under applicable law.  If additional permissions
+apply only to part of the Program, that part may be used separately
+under those permissions, but the entire Program remains governed by
+this License without regard to the additional permissions.
+
+  When you convey a copy of a covered work, you may at your option
+remove any additional permissions from that copy, or from any part of
+it.  (Additional permissions may be written to require their own
+removal in certain cases when you modify the work.)  You may place
+additional permissions on material, added by you to a covered work,
+for which you have or can give appropriate copyright permission.
+
+  Notwithstanding any other provision of this License, for material you
+add to a covered work, you may (if authorized by the copyright holders of
+that material) supplement the terms of this License with terms:
+
+    a) Disclaiming warranty or limiting liability differently from the
+    terms of sections 15 and 16 of this License; or
+
+    b) Requiring preservation of specified reasonable legal notices or
+    author attributions in that material or in the Appropriate Legal
+    Notices displayed by works containing it; or
+
+    c) Prohibiting misrepresentation of the origin of that material, or
+    requiring that modified versions of such material be marked in
+    reasonable ways as different from the original version; or
+
+    d) Limiting the use for publicity purposes of names of licensors or
+    authors of the material; or
+
+    e) Declining to grant rights under trademark law for use of some
+    trade names, trademarks, or service marks; or
+
+    f) Requiring indemnification of licensors and authors of that
+    material by anyone who conveys the material (or modified versions of
+    it) with contractual assumptions of liability to the recipient, for
+    any liability that these contractual assumptions directly impose on
+    those licensors and authors.
+
+  All other non-permissive additional terms are considered "further
+restrictions" within the meaning of section 10.  If the Program as you
+received it, or any part of it, contains a notice stating that it is
+governed by this License along with a term that is a further
+restriction, you may remove that term.  If a license document contains
+a further restriction but permits relicensing or conveying under this
+License, you may add to a covered work material governed by the terms
+of that license document, provided that the further restriction does
+not survive such relicensing or conveying.
+
+  If you add terms to a covered work in accord with this section, you
+must place, in the relevant source files, a statement of the
+additional terms that apply to those files, or a notice indicating
+where to find the applicable terms.
+
+  Additional terms, permissive or non-permissive, may be stated in the
+form of a separately written license, or stated as exceptions;
+the above requirements apply either way.
+
+  8. Termination.
+
+  You may not propagate or modify a covered work except as expressly
+provided under this License.  Any attempt otherwise to propagate or
+modify it is void, and will automatically terminate your rights under
+this License (including any patent licenses granted under the third
+paragraph of section 11).
+
+  However, if you cease all violation of this License, then your
+license from a particular copyright holder is reinstated (a)
+provisionally, unless and until the copyright holder explicitly and
+finally terminates your license, and (b) permanently, if the copyright
+holder fails to notify you of the violation by some reasonable means
+prior to 60 days after the cessation.
+
+  Moreover, your license from a particular copyright holder is
+reinstated permanently if the copyright holder notifies you of the
+violation by some reasonable means, this is the first time you have
+received notice of violation of this License (for any work) from that
+copyright holder, and you cure the violation prior to 30 days after
+your receipt of the notice.
+
+  Termination of your rights under this section does not terminate the
+licenses of parties who have received copies or rights from you under
+this License.  If your rights have been terminated and not permanently
+reinstated, you do not qualify to receive new licenses for the same
+material under section 10.
+
+  9. Acceptance Not Required for Having Copies.
+
+  You are not required to accept this License in order to receive or
+run a copy of the Program.  Ancillary propagation of a covered work
+occurring solely as a consequence of using peer-to-peer transmission
+to receive a copy likewise does not require acceptance.  However,
+nothing other than this License grants you permission to propagate or
+modify any covered work.  These actions infringe copyright if you do
+not accept this License.  Therefore, by modifying or propagating a
+covered work, you indicate your acceptance of this License to do so.
+
+  10. Automatic Licensing of Downstream Recipients.
+
+  Each time you convey a covered work, the recipient automatically
+receives a license from the original licensors, to run, modify and
+propagate that work, subject to this License.  You are not responsible
+for enforcing compliance by third parties with this License.
+
+  An "entity transaction" is a transaction transferring control of an
+organization, or substantially all assets of one, or subdividing an
+organization, or merging organizations.  If propagation of a covered
+work results from an entity transaction, each party to that
+transaction who receives a copy of the work also receives whatever
+licenses to the work the party's predecessor in interest had or could
+give under the previous paragraph, plus a right to possession of the
+Corresponding Source of the work from the predecessor in interest, if
+the predecessor has it or can get it with reasonable efforts.
+
+  You may not impose any further restrictions on the exercise of the
+rights granted or affirmed under this License.  For example, you may
+not impose a license fee, royalty, or other charge for exercise of
+rights granted under this License, and you may not initiate litigation
+(including a cross-claim or counterclaim in a lawsuit) alleging that
+any patent claim is infringed by making, using, selling, offering for
+sale, or importing the Program or any portion of it.
+
+  11. Patents.
+
+  A "contributor" is a copyright holder who authorizes use under this
+License of the Program or a work on which the Program is based.  The
+work thus licensed is called the contributor's "contributor version".
+
+  A contributor's "essential patent claims" are all patent claims
+owned or controlled by the contributor, whether already acquired or
+hereafter acquired, that would be infringed by some manner, permitted
+by this License, of making, using, or selling its contributor version,
+but do not include claims that would be infringed only as a
+consequence of further modification of the contributor version.  For
+purposes of this definition, "control" includes the right to grant
+patent sublicenses in a manner consistent with the requirements of
+this License.
+
+  Each contributor grants you a non-exclusive, worldwide, royalty-free
+patent license under the contributor's essential patent claims, to
+make, use, sell, offer for sale, import and otherwise run, modify and
+propagate the contents of its contributor version.
+
+  In the following three paragraphs, a "patent license" is any express
+agreement or commitment, however denominated, not to enforce a patent
+(such as an express permission to practice a patent or covenant not to
+sue for patent infringement).  To "grant" such a patent license to a
+party means to make such an agreement or commitment not to enforce a
+patent against the party.
+
+  If you convey a covered work, knowingly relying on a patent license,
+and the Corresponding Source of the work is not available for anyone
+to copy, free of charge and under the terms of this License, through a
+publicly available network server or other readily accessible means,
+then you must either (1) cause the Corresponding Source to be so
+available, or (2) arrange to deprive yourself of the benefit of the
+patent license for this particular work, or (3) arrange, in a manner
+consistent with the requirements of this License, to extend the patent
+license to downstream recipients.  "Knowingly relying" means you have
+actual knowledge that, but for the patent license, your conveying the
+covered work in a country, or your recipient's use of the covered work
+in a country, would infringe one or more identifiable patents in that
+country that you have reason to believe are valid.
+
+  If, pursuant to or in connection with a single transaction or
+arrangement, you convey, or propagate by procuring conveyance of, a
+covered work, and grant a patent license to some of the parties
+receiving the covered work authorizing them to use, propagate, modify
+or convey a specific copy of the covered work, then the patent license
+you grant is automatically extended to all recipients of the covered
+work and works based on it.
+
+  A patent license is "discriminatory" if it does not include within
+the scope of its coverage, prohibits the exercise of, or is
+conditioned on the non-exercise of one or more of the rights that are
+specifically granted under this License.  You may not convey a covered
+work if you are a party to an arrangement with a third party that is
+in the business of distributing software, under which you make payment
+to the third party based on the extent of your activity of conveying
+the work, and under which the third party grants, to any of the
+parties who would receive the covered work from you, a discriminatory
+patent license (a) in connection with copies of the covered work
+conveyed by you (or copies made from those copies), or (b) primarily
+for and in connection with specific products or compilations that
+contain the covered work, unless you entered into that arrangement,
+or that patent license was granted, prior to 28 March 2007.
+
+  Nothing in this License shall be construed as excluding or limiting
+any implied license or other defenses to infringement that may
+otherwise be available to you under applicable patent law.
+
+  12. No Surrender of Others' Freedom.
+
+  If conditions are imposed on you (whether by court order, agreement or
+otherwise) that contradict the conditions of this License, they do not
+excuse you from the conditions of this License.  If you cannot convey a
+covered work so as to satisfy simultaneously your obligations under this
+License and any other pertinent obligations, then as a consequence you may
+not convey it at all.  For example, if you agree to terms that obligate you
+to collect a royalty for further conveying from those to whom you convey
+the Program, the only way you could satisfy both those terms and this
+License would be to refrain entirely from conveying the Program.
+
+  13. Use with the GNU Affero General Public License.
+
+  Notwithstanding any other provision of this License, you have
+permission to link or combine any covered work with a work licensed
+under version 3 of the GNU Affero General Public License into a single
+combined work, and to convey the resulting work.  The terms of this
+License will continue to apply to the part which is the covered work,
+but the special requirements of the GNU Affero General Public License,
+section 13, concerning interaction through a network will apply to the
+combination as such.
+
+  14. Revised Versions of this License.
+
+  The Free Software Foundation may publish revised and/or new versions of
+the GNU General Public License from time to time.  Such new versions will
+be similar in spirit to the present version, but may differ in detail to
+address new problems or concerns.
+
+  Each version is given a distinguishing version number.  If the
+Program specifies that a certain numbered version of the GNU General
+Public License "or any later version" applies to it, you have the
+option of following the terms and conditions either of that numbered
+version or of any later version published by the Free Software
+Foundation.  If the Program does not specify a version number of the
+GNU General Public License, you may choose any version ever published
+by the Free Software Foundation.
+
+  If the Program specifies that a proxy can decide which future
+versions of the GNU General Public License can be used, that proxy's
+public statement of acceptance of a version permanently authorizes you
+to choose that version for the Program.
+
+  Later license versions may give you additional or different
+permissions.  However, no additional obligations are imposed on any
+author or copyright holder as a result of your choosing to follow a
+later version.
+
+  15. Disclaimer of Warranty.
+
+  THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
+APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
+HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
+OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
+THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
+PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
+IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
+ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
+
+  16. Limitation of Liability.
+
+  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
+WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
+THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
+GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
+USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
+DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
+PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
+EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
+SUCH DAMAGES.
+
+  17. Interpretation of Sections 15 and 16.
+
+  If the disclaimer of warranty and limitation of liability provided
+above cannot be given local legal effect according to their terms,
+reviewing courts shall apply local law that most closely approximates
+an absolute waiver of all civil liability in connection with the
+Program, unless a warranty or assumption of liability accompanies a
+copy of the Program in return for a fee.
+
+                     END OF TERMS AND CONDITIONS
+
+            How to Apply These Terms to Your New Programs
+
+  If you develop a new program, and you want it to be of the greatest
+possible use to the public, the best way to achieve this is to make it
+free software which everyone can redistribute and change under these terms.
+
+  To do so, attach the following notices to the program.  It is safest
+to attach them to the start of each source file to most effectively
+state the exclusion of warranty; and each file should have at least
+the "copyright" line and a pointer to where the full notice is found.
+
+    <one line to give the program's name and a brief idea of what it does.>
+    Copyright (C) <year>  <name of author>
+
+    This program is free software: you can redistribute it and/or modify
+    it under the terms of the GNU General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
+
+    This program is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU General Public License for more details.
+
+    You should have received a copy of the GNU General Public License
+    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+Also add information on how to contact you by electronic and paper mail.
+
+  If the program does terminal interaction, make it output a short
+notice like this when it starts in an interactive mode:
+
+    <program>  Copyright (C) <year>  <name of author>
+    This program comes with ABSOLUTELY NO WARRANTY; for details type `show w'.
+    This is free software, and you are welcome to redistribute it
+    under certain conditions; type `show c' for details.
+
+The hypothetical commands `show w' and `show c' should show the appropriate
+parts of the General Public License.  Of course, your program's commands
+might be different; for a GUI interface, you would use an "about box".
+
+  You should also get your employer (if you work as a programmer) or school,
+if any, to sign a "copyright disclaimer" for the program, if necessary.
+For more information on this, and how to apply and follow the GNU GPL, see
+<http://www.gnu.org/licenses/>.
+
+  The GNU General Public License does not permit incorporating your program
+into proprietary programs.  If your program is a subroutine library, you
+may consider it more useful to permit linking proprietary applications with
+the library.  If this is what you want to do, use the GNU Lesser General
+Public License instead of this License.  But first, please read
+<http://www.gnu.org/philosophy/why-not-lgpl.html>.

+ 50 - 0
Makefile

@@ -0,0 +1,50 @@
+#  cudadl version 0.8: Compute discrete logs in smooth group orders
+#                      using CUDA
+#  Copyright (C) 2012 by Ryan Henry and Ian Goldberg
+#                        {rhenry,iang}@cs.uwaterloo.ca
+#
+#  This program is free software: you can redistribute it and/or modify
+#  it under the terms of version 3 of the GNU General Public License as
+#  published by the Free Software Foundation.
+#
+#  This program is distributed in the hope that it will be useful,
+#  but WITHOUT ANY WARRANTY; without even the implied warranty of
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#  GNU General Public License for more details.
+#
+#  You should have received a copy of the GNU General Public License
+#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+CXXFLAGS=-g -Wall -O2
+NVCCOPTS=-g -arch sm_20 --ptxas-options=-v -O2
+NVCC=nvcc $(NVCCOPTS)
+
+CXXFILES = dlrho.cc gen_N.cc
+CUFILES = parrhoasm.cu dpstream.cu
+OFILES = dlrho.o gen_N.o cudadl.o
+TARGETS = gen_N dlrho
+
+WORDS = 24
+
+all: $(TARGETS)
+
+gen_N: gen_N.o
+	g++ -g -Wall $^ -o $@ -lntl -lgmp
+
+dlrho: dlrho.o cudadl.o
+	g++ -g -Wall $^ -o $@ -lntl -lgmp -L/usr/local/cuda/lib64 -lcudart
+
+dlrho.o: dlrho.cc
+	g++ $(CXXFLAGS) -I /usr/local/cuda/include $^ -c -o $@
+
+cudadl.o: parrhoasm.cu cios.asm dpstream.cu
+	$(NVCC) -c parrhoasm.cu -DWORDS=$(WORDS) -o cudadl.o
+
+cios.asm: gencios_reg_20
+	./gencios_reg_20 $(WORDS) > $@
+
+clean:
+	-rm -f $(OFILES)
+
+veryclean: clean
+	-rm -f $(TARGETS) cios.asm

+ 46 - 0
README

@@ -0,0 +1,46 @@
+cudadl-0.8
+23 Jan 2012
+Ryan Henry and Ian Goldberg
+{rhenry,iang}@cs.uwaterloo.ca
+http://crysp.uwaterloo.ca/software/
+
+This package contains the source code to our CUDA implementation of
+van Oorschot and Wiener's parallel version of the Pollard rho discrete
+log algorithm.  It is intended for use on 1536-bit moduli that are
+RSA numbers with smooth totient; that is, the modulus N=pq, where p and q
+are 768-bit primes, and the prime factors of p-1 and q-1 are all
+distinct and less then B, for a parameter B.  [The value 1536 is
+hardcoded as "WORDS = 24" (24*32*2 = 1536) in the Makefile; it is easy
+to change this value and recompile if desired.]  Note that this means
+the totient of N = \phi(N) = (p-1)(q-1) has all prime factors less than
+B; that is, \phi(n) is "B-smooth".
+
+Usage:
+
+1. Build the software.  You'll need:
+
+    NTL
+    GMP
+    NVIDIA CUDA Toolkit 3.1
+    2 M2050 (or other compute capability level 2.0) CUDA cards
+      [If you have more or just 1, you'll need to modify dlrho.cc,
+       unfortunately.]
+
+    Hopefully just typing "make" should work.  It will build gen_N and
+    dlrho.
+
+2. Create the modulus N as, for example, a 1536-bit RSA number whose
+   totient is 2^50-smooth:
+
+   ./gen_N 1536 50 > N
+
+3. Generate a DL problem mod N and solve it:
+
+    ./dlrho < N
+
+This software is described in "Solving Discrete Logarithms in
+Smooth-Order Groups with CUDA", CACR technical report 2012-02,
+http://www.cacr.math.uwaterloo.ca/techreports/2012/cacr2012-02.pdf
+
+This program is covered under version 3 of the GNU General Public
+Licence; see the file COPYING for more information.

+ 31 - 0
cudadl.h

@@ -0,0 +1,31 @@
+/*
+ *  cudadl version 0.8: Compute discrete logs in smooth group orders
+ *                      using CUDA
+ *  Copyright (C) 2012 by Ryan Henry and Ian Goldberg
+ *                        {rhenry,iang}@cs.uwaterloo.ca
+ *
+ *  This program is free software: you can redistribute it and/or modify
+ *  it under the terms of version 3 of the GNU General Public License as
+ *  published by the Free Software Foundation.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
+#ifndef __CUDADL_H__
+#define __CUDADL_H__
+
+#include <NTL/ZZ.h>
+#include <NTL/ZZ_p.h>
+
+NTL_CLIENT
+
+ZZ cuda_dl(const ZZ_p &base, const ZZ_p &target, const ZZ &order,
+		const ZZ &modulus);
+
+#endif

+ 256 - 0
dlrho.cc

@@ -0,0 +1,256 @@
+/*
+ *  cudadl version 0.8: Compute discrete logs in smooth group orders
+ *                      using CUDA
+ *  Copyright (C) 2012 by Ryan Henry and Ian Goldberg
+ *                        {rhenry,iang}@cs.uwaterloo.ca
+ *
+ *  This program is free software: you can redistribute it and/or modify
+ *  it under the terms of version 3 of the GNU General Public License as
+ *  published by the Free Software Foundation.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
+#include <fstream>
+#include <sys/time.h>
+#include <NTL/vec_ZZ.h>
+#include <NTL/ZZ_p.h>
+#include "cudadl.h"
+
+#include <sys/types.h>
+#include <sys/wait.h>
+#include <sys/socket.h>
+#include <cuda_runtime.h>
+
+NTL_CLIENT
+
+// Compute the discrete log of target mod p, to the given base.
+// p must be the current ZZ_p modulus.
+// Place the result in exp.  fvec is a vector of the factors of
+// (p-1)/2, which must each be small enough to compute discrete logs
+// with some other method (kangaroo, index calculus, GNFS, etc.).
+// label is "p" or "q", to be printed to report progress.
+// Return 0 on failure, 1 on success.
+static int p_dl(const ZZ_p &target, const ZZ_p &base, ZZ &exp,
+    const vec_ZZ &fvec, const string &label)
+{
+    const int flen = fvec.length();
+
+    // Compute phi(p)
+    ZZ phip;
+    phip = 2;
+    for (int i = 0; i < flen; ++i) {
+	phip *= fvec[i];
+    }
+
+    // Invariant: the desired exp \equiv (curexp mod curmodulus)
+    ZZ curmodulus, curexp;
+    curmodulus = 2;
+    curexp = 0;
+
+    for (int i = 0; i < flen; ++i) {
+	cout << label << " submodulus " << i+1 << " of " << flen << "... ";
+	cout.flush();
+	// Figure out exp mod fvec[i] by taking each side to the power
+	// of phirho/fvec[i] so that we're working in the
+	// subgroup of order fvec[i].
+	ZZ quotient = phip / fvec[i];
+	ZZ_p subgroup_base = power(base, quotient);
+	ZZ_p subgroup_target = power(target, quotient);
+
+	if (subgroup_base == 1) {
+	    // The original base wasn't a generator of the whole group.
+	    if (subgroup_target == 1) {
+		cout << "Non-unique solution (mod " << fvec[i] <<")\n";
+		continue;
+	    } else {
+		cout << "Target not in subgroup generated by base\n";
+		return 0;
+	    }
+	}
+
+	// Now use your favourite method to get the DL of
+	// subgroup_target with base subgroup_base, knowing that it's in
+	// the range [0,fvec[i]).
+
+	cout << "\n";
+	ZZ md = ZZ_p::modulus();
+	struct timeval st, et;
+	gettimeofday(&st, NULL);
+	ZZ subgroup_dl = cuda_dl(subgroup_base, subgroup_target, fvec[i],
+					md);
+	gettimeofday(&et, NULL);
+	unsigned long us_elapsed = (et.tv_sec-st.tv_sec)*1000000 +
+	    (et.tv_usec-st.tv_usec);
+	printf("%ld.%06ld seconds elapsed\n", us_elapsed/1000000,
+	    us_elapsed % 1000000);
+
+	CRT(curexp, curmodulus, subgroup_dl, fvec[i]);
+	// cout << "CRT\n";
+	// cout << "curexp = " << curexp << "\n";
+	// cout << "curmodulus = " << curmodulus << "\n\n";
+    }
+    // We'd like a non-negative answer back
+    if (curexp >= 0) {
+	exp = curexp;
+    } else {
+	exp = curexp + curmodulus;
+    }
+    return 1;
+}
+
+typedef struct {
+    pid_t pid;
+    int rfd;
+} PDLHandle;
+
+// Behave like p_dl, but do the work in an asynchronous subprocess
+static PDLHandle* p_dl_fork_start(const ZZ &p, const ZZ_p &target,
+    const ZZ_p &base, const vec_ZZ &fvec, const string &label, int deviceid)
+{
+    PDLHandle *handle = new PDLHandle;
+    int fds[2];
+    int res = socketpair(AF_UNIX, SOCK_STREAM, 0, fds);
+    if (res < 0) {
+	perror("socketpair");
+	delete handle;
+	return NULL;
+    }
+    handle->rfd = fds[0];
+    pid_t childpid = fork();
+    if (childpid == -1) {
+	perror("fork");
+	delete handle;
+	return NULL;
+    }
+    if (childpid == 0) {
+	// Child
+	close(fds[0]);
+	int wfd = fds[1];
+
+	ZZ exp;
+	ZZ_p::init(p);
+	ZZ_p target_p, base_p;
+	conv(target_p, rep(target));
+	conv(base_p, rep(base));
+
+	cudaError_t cudares = cudaSetDevice(deviceid);
+	if (cudares != cudaSuccess) {
+	    cerr << "Error setting CUDA device\n";
+	    exit(1);
+	}
+	int res = p_dl(target_p, base_p, exp, fvec, label);
+
+	if (res) {
+	    // Write the result back to the parent using wfd
+	    unsigned short explen = NumBytes(exp);
+	    res = write(wfd, &explen, sizeof(unsigned short));
+	    unsigned char expbuf[explen];
+	    BytesFromZZ(expbuf, exp, explen);
+	    res = write(wfd, expbuf, explen);
+	    close(wfd);
+	}
+
+	exit(0);
+    } else {
+	// Parent
+	close(fds[1]);
+    }
+
+    return handle;
+}
+
+// Wait until the subprocess started by p_dl_fork_start completes, and
+// return its result.  handle is cleaned up.
+static int p_dl_fork_join(PDLHandle *handle, ZZ &exp)
+{
+    if (!handle) return 0;
+    pid_t childpid = handle->pid;
+    unsigned short explen;
+    int res = read(handle->rfd, &explen, sizeof(unsigned short));
+    if (res < (int)sizeof(unsigned short)) {
+	close(handle->rfd);
+	delete handle;
+	waitpid(childpid, NULL, 0);
+	return 0;
+    }
+    unsigned char expbuf[explen];
+    res = read(handle->rfd, expbuf, explen);
+    if (res < explen) {
+	close(handle->rfd);
+	delete handle;
+	waitpid(childpid, NULL, 0);
+	return 0;
+    }
+    ZZFromBytes(exp, expbuf, explen);
+    close(handle->rfd);
+    delete handle;
+    waitpid(childpid, NULL, 0);
+    return 1;
+}
+
+int main(int argc, char **argv)
+{
+    // Initialize the prng with some randomness from the kernel
+    unsigned char randbuf[1024];
+    ifstream urand("/dev/urandom");
+    urand.read((char *)randbuf, sizeof(randbuf));
+    urand.close();
+    ZZ randzz = ZZFromBytes(randbuf, sizeof(randbuf));
+    SetSeed(randzz);
+
+    ZZ rho, p, q;
+    vec_ZZ pfvec, qfvec;
+    cin >> rho >> p >> pfvec >> q >> qfvec;
+
+    // Generate a DLP mod rho (in the large odd-order subgroup)
+    ZZ_p::init(rho);
+    ZZ_p base = power(random_ZZ_p(), 2);
+    ZZ_p target = power(random_ZZ_p(), 2);
+
+    cout << "base = " << base << "\n";
+    cout << "target = " << target << "\n";
+
+    ZZ exp_p, exp_q, exp;
+    int res_p = 0;
+    int res_q = 0;
+
+    PDLHandle *handle_p, *handle_q;
+
+    handle_p = p_dl_fork_start(p, target, base, pfvec, "p", 0);
+    handle_q = p_dl_fork_start(q, target, base, qfvec, "q", 1);
+    res_p = p_dl_fork_join(handle_p, exp_p);
+    res_q = p_dl_fork_join(handle_q, exp_q);
+
+    if (res_p && res_q) {
+	ZZ_p::init(rho);
+	ZZ pm1 = (p - 1)/2;
+	ZZ qm1 = (q - 1)/2;
+	if (exp_p < 0) exp_p += pm1;
+	if (exp_q < 0) exp_q += qm1;
+	CRT(exp_p, pm1, exp_q, qm1);
+	exp = exp_p;
+	// We'd like a non-negative answer back
+	if (exp < 0) exp += pm1;
+
+	cout << "exp = " << exp << "\n";
+	ZZ_p base_exp;
+	power(base_exp, base, exp);
+	if (base_exp == target) {
+	    cout << "CORRECT!\n";
+	} else {
+	    cout << "INCORRECT:\nbase^exp = " << base_exp << "\n";
+	    cout << "target   = " << target << "\n";
+	}
+    } else {
+	cout << "FAIL\n";
+    }
+
+    return 0;
+}

+ 121 - 0
dpstream.cu

@@ -0,0 +1,121 @@
+/*
+ *  cudadl version 0.8: Compute discrete logs in smooth group orders
+ *                      using CUDA
+ *  Copyright (C) 2012 by Ryan Henry and Ian Goldberg
+ *                        {rhenry,iang}@cs.uwaterloo.ca
+ *
+ *  This program is free software: you can redistribute it and/or modify
+ *  it under the terms of version 3 of the GNU General Public License as
+ *  published by the Free Software Foundation.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
+#ifndef __DPSTREAM_H__
+#define __DPSTREAM_H__
+
+#include <sm_11_atomic_functions.h>
+
+/* Routines to allow CUDA threads to write distinguished points into
+   a buffer that will be read by the host */
+
+__constant__ static unsigned int *DPbuffer = NULL;
+__constant__ static unsigned int DPbuffersize = 0;  // in words
+static unsigned int *DPbuffer_device = NULL;
+__device__ volatile static unsigned int DPbuffertail = 0;  // in words
+
+static const unsigned int DPrecordsize = (1+WORDS+4); // in words
+static unsigned int DPbufsize = 0;
+
+cudaError_t DPstreamInit(size_t numrecords)
+{
+    // Allocate the buffer
+    DPbufsize = DPrecordsize * numrecords;
+
+    if (cudaMalloc((void **)&DPbuffer_device, DPbufsize * sizeof(unsigned int))
+	    != cudaSuccess) {
+	return cudaErrorInitializationError;
+    }
+    // Zero it
+    cudaMemset(DPbuffer_device, 0, DPbufsize * sizeof(unsigned int));
+    // Initialize the device-visible pointers
+    unsigned int zero = 0;
+    cudaMemcpyToSymbol(DPbuffer, &DPbuffer_device, sizeof(unsigned int*));
+    cudaMemcpyToSymbol(DPbuffersize, &DPbufsize, sizeof(unsigned int));
+    cudaMemcpyToSymbol(DPbuffertail, &zero, sizeof(unsigned int));
+
+    return cudaSuccess;
+}
+
+void DPstreamEnd(void)
+{
+    if (DPbuffer_device) {
+	cudaFree(DPbuffer_device);
+    }
+    DPbuffer_device = NULL;
+}
+
+__device__ inline void write_dp(unsigned int*, unsigned long, unsigned long);
+
+// Call this from the device to write the following into the buffer:
+//   block and thread indices (1 word total)
+//   x (WORDS words)
+//   a (2 words)
+//   b (2 words)
+__device__ inline void DPstreamWrite(unsigned long a, unsigned long b)
+{
+    if (!DPbuffer) return;
+
+    // Atomically increment the DPbuffertail
+    if (DPbuffertail >= DPbuffersize) return;
+    unsigned int ouroffset = atomicAdd((unsigned int *)&DPbuffertail,
+		(unsigned int) (DPrecordsize));
+    if (ouroffset + DPrecordsize > DPbuffersize) {
+	atomicSub((unsigned int *)&DPbuffertail, (unsigned int)DPrecordsize);
+	return;
+    }
+    unsigned int *ourbuffer = DPbuffer + ouroffset;
+
+    write_dp(ourbuffer, a, b);
+}
+
+static void dpcallback(void *data, unsigned short threadId,
+    unsigned short blockId, string x, unsigned long a, unsigned long b);
+
+void DPstreamParse(void *data)
+{
+    // Get the number of words in the buffer
+    unsigned int bufsize;
+    cudaMemcpyFromSymbol(&bufsize, DPbuffertail, sizeof(unsigned int));
+    cerr << getpid() << " " << bufsize / DPrecordsize << " DPs\n";
+    unsigned int *dpbuf = (unsigned int*)calloc(bufsize, sizeof(unsigned int));
+    if (dpbuf) {
+	cudaMemcpy(dpbuf, DPbuffer_device, bufsize * sizeof(unsigned int),
+	    cudaMemcpyDeviceToHost);
+
+	for (unsigned int *dp=dpbuf; dp < dpbuf+bufsize; dp += DPrecordsize) {
+	    unsigned short threadId = (unsigned short)(dp[0]);
+	    unsigned short blockId = dp[0]>>16;
+	    unsigned long a = dp[WORDS+2];
+	    a <<= 32;
+	    a += dp[WORDS+1];
+	    unsigned long b = dp[WORDS+4];
+	    b <<= 32;
+	    b += dp[WORDS+3];
+	    string x((const char *)(dp+1), WORDS*sizeof(unsigned int));
+
+	    dpcallback(data, threadId, blockId, x, a, b);
+	}
+    }
+    free(dpbuf);
+    // Reset the buffer
+    unsigned int zero = 0;
+    cudaMemcpyToSymbol(DPbuffertail, &zero, sizeof(unsigned int));
+}
+#endif

+ 145 - 0
gen_N.cc

@@ -0,0 +1,145 @@
+/*
+ *  cudadl version 0.8: Compute discrete logs in smooth group orders
+ *                      using CUDA
+ *  Copyright (C) 2012 by Ryan Henry and Ian Goldberg
+ *                        {rhenry,iang}@cs.uwaterloo.ca
+ *
+ *  This program is free software: you can redistribute it and/or modify
+ *  it under the terms of version 3 of the GNU General Public License as
+ *  published by the Free Software Foundation.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
+#include <fstream>
+#include <NTL/vec_ZZ.h>
+#include <NTL/ZZ_p.h>
+
+NTL_CLIENT
+
+// Make one attempt at generating a "smooth" prime p; i.e. a prime p
+// such that all of the prime factors of p-1 are at most l_B bits long.
+// Return 1 on success (and p is set to the prime and fvec is set to the
+// list of factors of (p-1)/2).  Return 0 on failure.  It is expected that
+// this function will fail most of the time.  p will end up being
+// p_bits bits long.
+static int try_gen_smooth_prime(ZZ& p, vec_ZZ &fvec, long p_bits, long l_B)
+{
+    ZZ prod_so_far;
+    prod_so_far = 1;
+    fvec.SetLength(0);
+
+    while(NumBits(prod_so_far) < p_bits - 1) {
+	long next_bits = p_bits - 1 - NumBits(prod_so_far);
+	if (next_bits > l_B) next_bits = l_B;
+	// For the last prime, add a bit so we're not accidentally short
+	// one bit
+	if (next_bits < l_B) ++next_bits;
+	// There's only one prime of length < 3
+	if (next_bits < 3) next_bits = 3;
+
+	// cout << "next_bits = " << next_bits << "\n";
+	ZZ nextprime = GenPrime_ZZ(next_bits);
+	append(fvec, nextprime);
+	prod_so_far *= nextprime;
+	// cout << "next prime = " << nextprime << "\n";
+    }
+
+    // Now see if 2*prod_so_far - 1 is prime
+    p = prod_so_far * 2;
+    p += 1;
+    return NumBits(p) == p_bits && ProbPrime(p);
+}
+
+// Verify that all of the elements of v2 are coprime to (a) all of the
+// elements of v1, and (b) each other.  Return 1 if so, 0 if not.
+static int all_coprime(const vec_ZZ &v1, const vec_ZZ &v2)
+{
+    ZZ prod_so_far;
+    prod_so_far = 1;
+    const int v1l = v1.length();
+    for (int i=0; i < v1l; ++i) {
+	prod_so_far *= v1[i];
+    }
+
+    const int v2l = v2.length();
+    for (int i=0; i < v2l; ++i) {
+	ZZ nextnum = v2[i];
+	ZZ g = GCD(prod_so_far, nextnum);
+	if (g > 1) {
+	    return 0;
+	}
+	prod_so_far *= nextnum;
+    }
+    return 1;
+}
+
+// Generate rho, a product of two l_B-smooth primes.  rho should be
+// about rho_bits long.  p and q will be assigned the factors of rho.
+// pfvec and qfvec will be populated with the prime factors
+// of (p-1)/2 and (q-1)/2 respectively.  If strong==1, then 4*rho+1 must
+// also be prime.  [p and q will be 2 mod 3, so 2*p*q+1 will be
+// divisible by 3 always.]
+static void gen_rho(ZZ &rho, vec_ZZ &pfvec, vec_ZZ &qfvec, ZZ &p, ZZ &q,
+	long rho_bits, long l_B, int strong)
+{
+    do {
+	pfvec.SetLength(0);
+	qfvec.SetLength(0);
+	while(1) {
+	    cerr << ".";
+	    cerr.flush();
+	    int ret = try_gen_smooth_prime(p, pfvec, rho_bits / 2, l_B);
+	    if (ret == 1 && all_coprime(qfvec /* qfvec is empty */, pfvec)) break;
+	}
+	cerr << "\np = " << p << "\n\n";
+	while(1) {
+	    cerr << ".";
+	    cerr.flush();
+	    int ret = try_gen_smooth_prime(q, qfvec, rho_bits - NumBits(p), l_B);
+	    if (ret == 1 && all_coprime(pfvec, qfvec)) break;
+	}
+	cerr << "\nq = " << q << "\n\n";
+	rho = p * q;
+	if (!strong) break;
+	ZZ R = 4*rho + 1;
+	if (!ProbPrime(R)) {
+	    cerr << "4*rho + 1  = " << R << " not prime.\n\n";
+	} else {
+	    cerr << "R = " << R << "\n\n";
+	    break;
+	}
+    } while (1);
+}
+
+
+int main(int argc, char **argv)
+{
+    long rho_bits = argc > 1 ? atoi(argv[1]) : 1536;
+    long l_B = argc > 2 ? atoi(argv[2]) : 30;
+    int strong = argc > 3 ? atoi(argv[3]) : 0;
+
+    // Initialize the prng with some randomness from the kernel
+    unsigned char randbuf[1024];
+    ifstream urand("/dev/urandom");
+    urand.read((char *)randbuf, sizeof(randbuf));
+    urand.close();
+    ZZ randzz = ZZFromBytes(randbuf, sizeof(randbuf));
+    SetSeed(randzz);
+
+    ZZ rho, p, q;
+    vec_ZZ pfvec, qfvec;
+
+    gen_rho(rho, pfvec, qfvec, p, q, rho_bits, l_B, strong);
+
+    cout << rho << "\n" << p << "\n" << pfvec << "\n" << q << "\n"
+	    << qfvec << "\n";
+
+    return 0;
+}

+ 323 - 0
gencios_reg_20

@@ -0,0 +1,323 @@
+#!/usr/bin/perl
+
+#  cudadl version 0.8: Compute discrete logs in smooth group orders
+#                      using CUDA
+#  Copyright (C) 2012 by Ryan Henry and Ian Goldberg
+#                        {rhenry,iang}@cs.uwaterloo.ca
+#
+#  This program is free software: you can redistribute it and/or modify
+#  it under the terms of version 3 of the GNU General Public License as
+#  published by the Free Software Foundation.
+#
+#  This program is distributed in the hope that it will be useful,
+#  but WITHOUT ANY WARRANTY; without even the implied warranty of
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#  GNU General Public License for more details.
+#
+#  You should have received a copy of the GNU General Public License
+#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+use strict;
+
+die "Usage: $0 WORDS\n" unless $#ARGV == 0;
+
+my $WORDS = shift;
+my $WORDSm1 = ($WORDS-1);
+my $WORDSp1 = ($WORDS+1);
+my $WORDSp2 = ($WORDS+2);
+
+my $i;
+
+## Output an evil hack that declares a bunch of registers in the module
+## scope
+
+print <<EOA;
+// An evil hack that declares a bunch of registers in the module scope
+__device__ __noinline__ void evilhack() {
+    asm("ret;}\\n"
+        ".reg.u32 \$xr<$WORDSp1>;\\n"
+        ".reg.u32 \$cg<$WORDS>;\\n"
+        ".reg.u32 \$cy<$WORDS>;\\n"
+	".reg.u32 \$yi,\$crp,\$m;\\n"
+	".reg.u64 \$cs,\$ct;\\n"
+	".reg.u32 \$zr<$WORDSp2>;\\n"
+	".reg.u32 \$crho<$WORDS>;\\n"
+	".func hackend {\\n"
+EOA
+print "\t\"ret.uni;\\n\"\n";
+
+print "    );\n";
+print "}\n";
+
+print <<'EOA';
+__device__ inline void loadx(unsigned int *x)
+{
+    asm(
+EOA
+
+for ($i=0;$i<$WORDS;++$i) {
+    my $xoff = $i*4;
+    print "\t\"ld.global.u32 \$xr$i, [%[x]+$xoff];\\n\"\n";
+}
+print "\t\"// Load c_rho into registers\\n\"\n";
+my ($xoff, $xoffm1);
+my $croff;
+for($i=0;$i<$WORDS;++$i) {
+    $croff = $i*4;
+    print "\t\"ld.const.u32 \$crho$i, [c_rho+$croff];\\n\"\n";
+}
+print "\t\"mov.u32 \$crp, %[crp];\\n\"\n";
+print "\t\"// Load c_g into registers\\n\"\n";
+for($i=0;$i<$WORDS;++$i) {
+    $croff = $i*4;
+    print "\t\"ld.const.u32 \$cg$i, [c_g+$croff];\\n\"\n";
+}
+print "\t\"// Load c_y into registers\\n\"\n";
+for($i=0;$i<$WORDS;++$i) {
+    $croff = $i*4;
+    print "\t\"ld.const.u32 \$cy$i, [c_y+$croff];\\n\"\n";
+}
+
+
+print <<'EOA';
+    : : [x] "l" (x), [crp] "r" (c_rho_prime));
+}
+
+__device__ inline void savex(unsigned int *x)
+{
+    asm(
+EOA
+
+for ($i=0;$i<$WORDS;++$i) {
+    my $xoff = $i*4;
+    print "\t\"st.global.u32 [%[x]+$xoff], \$xr$i;\\n\"\n";
+}
+
+print <<'EOA';
+    : : [x] "l" (x));
+}
+EOA
+
+print <<'EOA';
+__device__ void modmul(int multtype)
+// x <- (x * y) * r_inv   (mod rho);
+// where y is (multtype == 0 ? c_g : multtype == 1 ? c_y : x)
+{
+    asm(
+    ".reg.pred $mtp0,$mtp1;\n"
+EOA
+print "    \".reg.pred \$retaddr<", $WORDS-2, ">;\\n\"\n";
+print <<'EOA';
+    "mov.u32 $m, 0;\n"
+    "setp.eq.u32 $mtp0, %[tp], $m;\n"
+    "mov.u32 $m, 1;\n"
+    "setp.eq.u32 $mtp1, %[tp], $m;\n"
+    : : [tp] "r" (multtype));
+EOA
+
+print "    asm(\n";
+print "\t\"// First iteration: i=0 and z is unset\\n\"\n";
+
+# Pick the correct value of $yi
+print "\t\"selp.u32 \$yi, \$cg0, \$xr0, \$mtp0;\\n\"\n";
+print "\t\"selp.u32 \$yi, \$cy0, \$yi, \$mtp1;\\n\"\n";
+
+print "\t\"mul.wide.u32 \$cs, \$xr0, \$yi;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr0, \$cs;\\n\"\n";
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+
+for($i=1; $i<$WORDS; ++$i) {
+    $xoff = $i*32*4;
+    print "\t\"mad.wide.u32 \$cs, \$xr$i, \$yi, \$cs;\\n\"\n";
+    print "\t\"cvt.u32.u64 \$zr$i, \$cs;\\n\"\n";
+    print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+}
+print "\t\"cvt.u32.u64 \$zr$WORDS, \$cs;\\n\"\n";
+
+print "\t\"mul.lo.u32 \$m, \$zr0, \$crp;\\n\"\n";
+print "\t\"cvt.u64.u32 \$ct, \$zr0;\\n\"\n";
+print "\t\"mad.wide.u32 \$cs, \$m, \$crho0, \$ct;\\n\"\n";
+
+for($i=1;$i<$WORDS;++$i) {
+    my $im1 = ($i-1);
+    print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+    print "\t\"cvt.u64.u32 \$ct, \$zr$i;\\n\"\n";
+    print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+    print "\t\"mad.wide.u32 \$cs, \$m, \$crho$i, \$cs;\\n\"\n";
+    print "\t\"cvt.u32.u64 \$zr$im1, \$cs;\\n\"\n";
+}
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+print "\t\"cvt.u64.u32 \$ct, \$zr$WORDS;\\n\"\n";
+print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr$WORDSm1, \$cs;\\n\"\n";
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr$WORDS, \$cs;\\n\"\n";
+
+my $j;
+for ($j=1; $j<$WORDS-1; ++$j) {
+
+print <<'EOA';
+	"// Middle iterations\n"
+EOA
+
+# Pick the correct value of $yi
+my $yioff = $j*4;
+print "\t\"selp.u32 \$yi, \$cg$j, \$xr$j, \$mtp0;\\n\"\n";
+print "\t\"selp.u32 \$yi, \$cy$j, \$yi, \$mtp1;\\n\"\n";
+my $jm1 = $j-1;
+my $jm2 = $j-2;
+if ($j==1) {
+    print "\t\"setp.eq.u32 \$retaddr0,\$xr0,\$xr0;\\n\"\n";
+} else {
+    print "\t\"setp.eq.u32 \$retaddr$jm1|\$retaddr$jm2,\$xr0,\$xr0;\\n\"\n";
+}
+print "\t\"bra.uni innerloop;\\n\"\n";
+print "\t\"innerret$jm1:\\n\"\n";
+}
+
+print <<'EOA';
+	"// Last iteration: store the output in x instead of z\n"
+EOA
+
+# Pick the correct value of $yi
+my $yioff = ($WORDSm1)*4;
+print "\t\"selp.u32 \$yi, \$cg$WORDSm1, \$xr$WORDSm1, \$mtp0;\\n\"\n";
+print "\t\"selp.u32 \$yi, \$cy$WORDSm1, \$yi, \$mtp1;\\n\"\n";
+
+print "\t\"cvt.u64.u32 \$ct, \$zr0;\\n\"\n";
+print "\t\"mad.wide.u32 \$cs, \$xr0, \$yi, \$ct;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr0, \$cs;\\n\"\n";
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+
+for($i=1; $i<$WORDS; ++$i) {
+    $xoff = $i*32*4;
+    print "\t\"cvt.u64.u32 \$ct, \$zr$i;\\n\"\n";
+    print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+    print "\t\"mad.wide.u32 \$cs, \$xr$i, \$yi, \$cs;\\n\"\n";
+    print "\t\"cvt.u32.u64 \$zr$i, \$cs;\\n\"\n";
+    print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+}
+print "\t\"cvt.u64.u32 \$ct, \$zr$WORDS;\\n\"\n";
+print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr$WORDS, \$cs;\\n\"\n";
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr$WORDSp1, \$cs;\\n\"\n";
+
+print "\t\"mul.lo.u32 \$m, \$zr0, \$crp;\\n\"\n";
+print "\t\"cvt.u64.u32 \$ct, \$zr0;\\n\"\n";
+print "\t\"mad.wide.u32 \$cs, \$m, \$crho0, \$ct;\\n\"\n";
+
+for($i=1;$i<$WORDS;++$i) {
+    my $im1 = $i-1;
+    print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+    print "\t\"cvt.u64.u32 \$ct, \$zr$i;\\n\"\n";
+    print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+    print "\t\"mad.wide.u32 \$cs, \$m, \$crho$i, \$cs;\\n\"\n";
+    print "\t\"cvt.u32.u64 \$xr$im1, \$cs;\\n\"\n";
+}
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+print "\t\"cvt.u64.u32 \$ct, \$zr$WORDS;\\n\"\n";
+print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+print "\t\"cvt.u32.u64 \$xr$WORDSm1, \$cs;\\n\"\n";
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+print "\t\"cvt.u64.u32 \$ct, \$zr$WORDSp1;\\n\"\n";
+print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+print "\t\"cvt.u32.u64 \$xr$WORDS, \$cs;\\n\"\n";
+
+print <<'EOA';
+        );
+EOA
+
+# If x >= c_rho, x <- x - c_rho
+
+print "    asm(\n";
+print "\t\".reg.pred \$gtP,\$ltP;\\n\"\n";
+print "\t\"mov.u32 \$m, 0;\\n\"\n";
+print "\t\"setp.gt.u32 \$gtP, \$xr$WORDS, \$m;\\n\"\n";
+print "\t\"\@\$gtP bra \$comp_done;\\n\"\n";
+for ($i = $WORDS-1; $i >= 1; --$i) {
+    print "\t\"setp.gt.u32 \$gtP, \$xr$i, \$crho$i;\\n\"\n";
+    print "\t\"\@\$gtP bra \$comp_done;\\n\"\n";
+    print "\t\"setp.lt.u32 \$ltP, \$xr$i, \$crho$i;\\n\"\n";
+    print "\t\"\@\$ltP bra \$comp_done;\\n\"\n";
+}
+print "\t\"setp.ge.u32 \$gtP, \$xr0, \$crho0;\\n\"\n";
+print "\t\"\$comp_done:\\n\"\n";
+
+# Do the subtraction
+print "\t\"@\$gtP sub.cc.u32 \$xr0, \$xr0, \$crho0;\\n\"\n";
+for ($i=1; $i<$WORDS; ++$i) {
+    print "\t\"@\$gtP subc.cc.u32 \$xr$i, \$xr$i, \$crho$i;\\n\"\n";
+}
+print "\t\"@\$gtP subc.u32 \$xr$WORDS, \$xr$WORDS, \$m;\\n\"\n";
+
+print "\t\"bra.uni mulmoddone;\\n\"\n";
+
+print "\t\"innerloop:\\n\"\n";
+print "\t\"cvt.u64.u32 \$ct, \$zr0;\\n\"\n";
+print "\t\"mad.wide.u32 \$cs, \$xr0, \$yi, \$ct;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr0, \$cs;\\n\"\n";
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+
+my $i;
+for($i=1; $i<$WORDS; ++$i) {
+    print "\t\"cvt.u64.u32 \$ct, \$zr$i;\\n\"\n";
+    print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+    print "\t\"mad.wide.u32 \$cs, \$xr$i, \$yi, \$cs;\\n\"\n";
+    print "\t\"cvt.u32.u64 \$zr$i, \$cs;\\n\"\n";
+    print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+}
+print "\t\"cvt.u64.u32 \$ct, \$zr$WORDS;\\n\"\n";
+print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr$WORDS, \$cs;\\n\"\n";
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr$WORDSp1, \$cs;\\n\"\n";
+
+print "\t\"mul.lo.u32 \$m, \$zr0, \$crp;\\n\"\n";
+print "\t\"cvt.u64.u32 \$ct, \$zr0;\\n\"\n";
+print "\t\"mad.wide.u32 \$cs, \$m, \$crho0, \$ct;\\n\"\n";
+
+for($i=1;$i<$WORDS;++$i) {
+    my $im1 = ($i-1);
+    print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+    print "\t\"cvt.u64.u32 \$ct, \$zr$i;\\n\"\n";
+    print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+    print "\t\"mad.wide.u32 \$cs, \$m, \$crho$i, \$cs;\\n\"\n";
+    print "\t\"cvt.u32.u64 \$zr$im1, \$cs;\\n\"\n";
+}
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+print "\t\"cvt.u64.u32 \$ct, \$zr$WORDS;\\n\"\n";
+print "\t\"add.u64 \$cs, \$cs, \$ct;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr$WORDSm1, \$cs;\\n\"\n";
+print "\t\"shr.u64 \$cs, \$cs, 32;\\n\"\n";
+print "\t\"cvt.u32.u64 \$zr$WORDS, \$cs;\\n\"\n";
+
+for ($i=0;$i<$WORDS-2;++$i) {
+    print "\t\"\@\$retaddr$i bra.uni innerret$i;\\n\"\n";
+}
+print "\t\"mulmoddone:\\n\"\n";
+
+print "    );\n";
+print "}\n";
+
+print <<'EOA';
+
+__device__ inline void write_dp(unsigned int *buf, unsigned long a, unsigned long b)
+{
+    unsigned short blockid = blockIdx.x + gridDim.x*blockIdx.y;
+    unsigned short threadid = threadIdx.x + blockDim.x*threadIdx.y + blockDim.x*blockDim.y*threadIdx.z;
+    buf[0] = (((unsigned int)blockid) << 16) + threadid;
+    asm(
+EOA
+for($i=0;$i<$WORDS;++$i) {
+    my $off = 4*($i+1);
+    print "\t\"st.global.u32 [%0+$off], \$xr$i;\\n\"\n";
+}
+print <<'EOA';
+    : : "l" (buf));
+    buf[WORDS+1] = (unsigned int)a;
+    buf[WORDS+2] = a>>32;
+    buf[WORDS+3] = (unsigned int)b;
+    buf[WORDS+4] = b>>32;
+}
+EOA

+ 601 - 0
parrhoasm.cu

@@ -0,0 +1,601 @@
+/*
+ *  cudadl version 0.8: Compute discrete logs in smooth group orders
+ *                      using CUDA
+ *  Copyright (C) 2012 by Ryan Henry and Ian Goldberg
+ *                        {rhenry,iang}@cs.uwaterloo.ca
+ *
+ *  This program is free software: you can redistribute it and/or modify
+ *  it under the terms of version 3 of the GNU General Public License as
+ *  published by the Free Software Foundation.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
+#include "cudadl.h"
+#include <fstream>
+#include <NTL/vec_ZZ.h>
+#include <string.h>
+#include <limits.h>
+#include <sys/time.h>
+#include <time.h>
+#include <string>
+#include <sstream>
+#include <utility>
+#include <map>
+
+// #define CHECK_RESULTS
+
+// #include "cuPrintf.cu"
+
+#define BITS_PER_WORD (8 * sizeof(unsigned int))
+#define TWO_32_DIV_3 1431655765U // floor( 2^32 / 3 )
+#define TWO_32_DIV_3_X2 2863311530U // 2 * floor( 2^32 / 3 )
+#define TWO_64_DIV_3 6148914691236517205UL // floor( 2^64 / 3 )
+
+#define SIZE_G WORDS
+#define SIZE_Y WORDS
+#define SIZE_X (WORDS + 1)
+
+#include "dpstream.cu"
+
+// #define X(idx) x[idx<<5]
+// #define Y(idx) (multtype == 0 ? c_g[idx] : multtype == 1 ? c_y[idx] : x[idx<<5])
+// #define Y(idx) ((c_g[idx]&typemask0) | (c_y[idx]&typemask1) | (x[idx<<5]&typemask2))
+// #define Z(idx) z[idx<<5]
+
+NTL_CLIENT
+
+#ifdef CHECK_RESULTS
+static void dump(const char *prefix, const unsigned int *di, size_t words)
+{
+    size_t l = words * 4;
+    const unsigned char *d = (const unsigned char *) di;
+    d += l;
+    printf("%s=", prefix);
+    while(l)
+    {
+        --d;
+        printf("%02X", *d);
+        --l;
+    }
+    printf("\n");
+}
+#endif
+
+static void checkCUDAError(const char *msg)
+{
+    cudaError_t err = cudaGetLastError();
+    if (cudaSuccess != err) 
+    {
+        fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString(err));
+        exit(EXIT_FAILURE);
+    }                         
+}
+
+typedef struct
+{
+    unsigned int x[SIZE_X];
+    unsigned int a[2], b[2];
+} GlobalThreadState;
+
+
+__device__ __constant__ unsigned int c_rho[WORDS + 1], c_r_inv[WORDS + 1], c_rho_prime;
+__device__ __constant__ unsigned int c_y[SIZE_Y], c_g[SIZE_G];
+
+/*
+__device__ unsigned int getgy(int i)
+{
+    return c_rho[i] + c_y[i] + c_g[i];
+}
+*/
+
+#include "cios.asm"
+
+#if 0
+__device__ void _sub(unsigned int *x)
+// x <- x - y
+{
+    asm("sub.cc.u32  %0, %1, %2;" : "=r"(X(0))  : "r"(X(0)),  "r"(c_rho[0]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(1))  : "r"(X(1)),  "r"(c_rho[1]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(2))  : "r"(X(2)),  "r"(c_rho[2]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(3))  : "r"(X(3)),  "r"(c_rho[3]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(4))  : "r"(X(4)),  "r"(c_rho[4]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(5))  : "r"(X(5)),  "r"(c_rho[5]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(6))  : "r"(X(6)),  "r"(c_rho[6]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(7))  : "r"(X(7)),  "r"(c_rho[7]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(8))  : "r"(X(8)),  "r"(c_rho[8]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(9))  : "r"(X(9)),  "r"(c_rho[9]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(10)) : "r"(X(10)), "r"(c_rho[10]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(11)) : "r"(X(11)), "r"(c_rho[11]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(12)) : "r"(X(12)), "r"(c_rho[12]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(13)) : "r"(X(13)), "r"(c_rho[13]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(14)) : "r"(X(14)), "r"(c_rho[14]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(15)) : "r"(X(15)), "r"(c_rho[15]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(16)) : "r"(X(16)), "r"(c_rho[16]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(17)) : "r"(X(17)), "r"(c_rho[17]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(18)) : "r"(X(18)), "r"(c_rho[18]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(19)) : "r"(X(19)), "r"(c_rho[19]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(20)) : "r"(X(20)), "r"(c_rho[20]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(21)) : "r"(X(21)), "r"(c_rho[21]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(22)) : "r"(X(22)), "r"(c_rho[22]));
+    asm("subc.cc.u32 %0, %1, %2;" : "=r"(X(23)) : "r"(X(23)), "r"(c_rho[23]));
+    asm("subc.u32    %0, %1, %2;" : "=r"(X(24)) : "r"(X(24)), "r"(c_rho[24]));
+}
+
+__device__ inline bool _gt(const unsigned int * x)
+// returns true iff x > y  ;  x is of length WORDS+1 and y is of length WORDS
+{
+    if (X(24)) return true;
+    if (X(23) < c_rho[23]) return false;
+    if (X(23) > c_rho[23]) return true;
+    if (X(22) < c_rho[22]) return false;
+    if (X(22) > c_rho[22]) return true;
+    if (X(21) < c_rho[21]) return false;
+    if (X(21) > c_rho[21]) return true;
+    if (X(20) < c_rho[20]) return false;
+    if (X(20) > c_rho[20]) return true;
+    if (X(19) < c_rho[19]) return false;
+    if (X(19) > c_rho[19]) return true;
+    if (X(18) < c_rho[18]) return false;
+    if (X(18) > c_rho[18]) return true;
+    if (X(17) < c_rho[17]) return false;
+    if (X(17) > c_rho[17]) return true;
+    if (X(16) < c_rho[16]) return false;
+    if (X(16) > c_rho[16]) return true;
+    if (X(15) < c_rho[15]) return false;
+    if (X(15) > c_rho[15]) return true;
+    if (X(14) < c_rho[14]) return false;
+    if (X(14) > c_rho[14]) return true;
+    if (X(13) < c_rho[13]) return false;
+    if (X(13) > c_rho[13]) return true;
+    if (X(12) < c_rho[12]) return false;
+    if (X(12) > c_rho[12]) return true;
+    if (X(11) < c_rho[11]) return false;
+    if (X(11) > c_rho[11]) return true;
+    if (X(10) < c_rho[10]) return false;
+    if (X(10) > c_rho[10]) return true;
+    if (X(9)  < c_rho[9])  return false;
+    if (X(9)  > c_rho[9])  return true;
+    if (X(8)  < c_rho[8])  return false;
+    if (X(8)  > c_rho[8])  return true;
+    if (X(7)  < c_rho[7])  return false;
+    if (X(7)  > c_rho[7])  return true;
+    if (X(6)  < c_rho[6])  return false;
+    if (X(6)  > c_rho[6])  return true;
+    if (X(5)  < c_rho[5])  return false;
+    if (X(5)  > c_rho[5])  return true;
+    if (X(4)  < c_rho[4])  return false;
+    if (X(4)  > c_rho[4])  return true;
+    if (X(3)  < c_rho[3])  return false;
+    if (X(3)  > c_rho[3])  return true;
+    if (X(2)  < c_rho[2])  return false;
+    if (X(2)  > c_rho[2])  return true;
+    if (X(1)  < c_rho[1])  return false;
+    if (X(1)  > c_rho[1])  return true;
+    return (X(0) > c_rho[0]);
+}
+#endif
+
+#define nmult 1000
+#define nlaunch 2
+
+#define threadDimx 32
+
+__global__ void cudaMulmod(GlobalThreadState *global_ts, unsigned long long order)
+// x <- x * y mod rho
+{
+    // extern __shared__ unsigned int sharedmem[];
+    register int i;
+    register int tid = threadIdx.x + threadDimx * threadIdx.y;
+    register int global_tid = tid + blockDim.x * blockDim.y * blockIdx.x;
+    unsigned long long a, b;
+    unsigned int *global_x_base = global_ts[global_tid].x;
+    unsigned int *global_a_base = global_ts[global_tid].a;
+    unsigned int *global_b_base = global_ts[global_tid].b;
+    // unsigned int *shared_table_base = sharedmem;
+    // unsigned int *shared_x_base = shared_table_base + threadIdx.y * 32 * (SIZE_X) + threadIdx.x;
+    unsigned int xlow;
+
+    // cuPrintf("d_z = %08X%08X%08X\n", global_x_base[2], global_x_base[1], global_x_base[0]);
+    loadx(global_x_base);
+    a = global_a_base[1];
+    a <<= 32;
+    a |= global_a_base[0];
+    b = global_b_base[1];
+    b <<= 32;
+    b |= global_b_base[0];
+    // cuPrintf("s_A = %08X\n", (unsigned int)a);
+    // cuPrintf("s_B = %08X\n", (unsigned int)b);
+    // cuPrintf("s_a = %08X%08X\n", global_a_base[1], global_a_base[0]);
+    // cuPrintf("s_b = %08X%08X\n", global_b_base[1], global_b_base[0]);
+    for (i = 0; i < nmult; ++i)
+    {
+        //memset(ts[tid].z, 0, (WORDS + 2) * sizeof(unsigned int));
+        //__syncthreads();
+        //cuPrintf("x    = %p %08X%08X%08X\n", ts[tid].x, ts[tid].x[2], ts[tid].x[1], ts[tid].x[0]);
+        //cuPrintf("y    = %p %08X%08X\n", c_y, c_y[1], c_y[0]);
+        //cuPrintf("z    = %p %08X%08X%08X%08X\n", ts[tid].z, ts[tid].z[3], ts[tid].z[2],  ts[tid].z[1], ts[tid].z[0]);
+        int multtype;
+	asm("mov.u32 %0, $xr0;" : "=r" (xlow));
+        if (xlow < TWO_32_DIV_3)
+        {
+            multtype = 0;
+	    a += 1;
+	    // cuPrintf("inc a\n");
+        }
+        else if (xlow < TWO_32_DIV_3_X2)
+        {
+            multtype = 1;
+	    b += 1;
+	    // cuPrintf("inc b\n");
+        }
+        else
+        {
+            multtype = 2;
+	    a += a;
+	    b += b;
+	    // cuPrintf("double\n");
+        }
+	if (a > order) a -= order;
+	if (b > order) b -= order;
+        modmul(multtype);
+        //memcpy(ts[tid].x, ts[tid].z, (WORDS + 1) * sizeof(unsigned int));
+	/*
+        if (_gt(shared_x_base)) {
+            _sub(shared_x_base);
+	}
+	*/
+        //cuPrintf("y    = %p %08X%08X\n", c_y, c_y[1], c_y[0]);
+        //cuPrintf("z    = %p %08X%08X\n", ts[tid].x, ts[tid].x[1], ts[tid].x[0]);
+        //cuPrintf("y    = %p %08X%08X\n", c_y, c_y[1], c_y[0]);
+
+	// Check for a distinguished point
+	asm("mov.u32 %0, $xr0;" : "=r" (xlow));
+	if ((xlow & 0x000003ff) == 0 || order < (1<<20)) {
+	    DPstreamWrite(a,b);
+	}
+    }
+    savex(global_x_base);
+    global_a_base[1] = (a>>32);
+    global_a_base[0] = (unsigned int)a;
+    global_b_base[1] = (b>>32);
+    global_b_base[0] = (unsigned int)b;
+    // cuPrintf("d_z = %08X%08X%08X\n", global_x_base[2], global_x_base[1], global_x_base[0]);
+    // cuPrintf("d_A = %08X\n", (unsigned int)a);
+    // cuPrintf("d_B = %08X\n", (unsigned int)b);
+    // cuPrintf("d_a = %08X%08X\n", global_a_base[1], global_a_base[0]);
+    // cuPrintf("d_b = %08X%08X\n", global_b_base[1], global_b_base[0]);
+    if (!order) evilhack();
+}
+
+int nthreads = 25600;
+int nblocks = 50;
+
+typedef map<std::string, pair<ZZ,ZZ> > DTable;
+
+struct CBData {
+    const ZZ_p &base;
+    const ZZ_p &target;
+    const ZZ &order;
+    unsigned long long numdp;
+    DTable dtable;
+    bool found_collision;
+    ZZ expon;
+
+    CBData(const ZZ_p &_base, const ZZ_p &_target, const ZZ &_order) :
+	base(_base), target(_target), order(_order), numdp(0),
+	found_collision(false) {}
+};
+
+static void dpcallback(void *cbdata, unsigned short threadId,
+    unsigned short blockId, string x, unsigned long a, unsigned long b)
+{
+    CBData *d = (CBData*)cbdata;
+    // WARNING: this assumes
+    //    sizeof(unsigned long) == sizeof(unsigned long long) !
+    ZZ zz_a = to_ZZ((unsigned long)a);
+    ZZ zz_b = to_ZZ((unsigned long)b);
+
+    //ZZ_p dp = power(d->base, zz_a) * power(d->target, zz_b);
+
+    // cerr << "DP " << ++(d->numdp) << "\r";
+    pair<ZZ,ZZ> ab(zz_a,zz_b);
+    pair<DTable::iterator, bool> res;
+    res = d->dtable.insert(DTable::value_type(x, ab));
+    if (!res.second) {
+	// Collision!
+	ZZ adiff = res.first->second.first - a;
+	ZZ bdiff = b - res.first->second.second;
+	if (bdiff < 0) bdiff += d->order;
+	ZZ binv;
+	if (InvModStatus(binv, bdiff, d->order) == 0) {
+	    d->expon = MulMod(binv, adiff, d->order);
+	    d->found_collision = true;
+	} else {
+	    if (d->order > (1<<20)) {
+		cerr << "Unhelpful collision\n";
+	    }
+	}
+    }
+}
+
+ZZ cuda_dl(const ZZ_p &base, const ZZ_p &target, const ZZ &order,
+		const ZZ &modulus)
+{
+    unsigned long long totmicros = 0;
+    ZZ_pBak pbak;
+    pbak.save();
+
+    if (NumBits(modulus) <= ((WORDS-1)*BITS_PER_WORD) ||
+	    NumBits(modulus) > WORDS*BITS_PER_WORD) {
+	cerr << "modulus is not " << WORDS << " words long.\n";
+	exit(1);
+    }
+
+    long orderbits = NumBits(order);
+    if (orderbits > 60) {
+	cerr << "order is larger than 60 bits.\n";
+	exit(1);
+    }
+
+    unsigned long long orderll = trunc_long(order, 32) +
+	((unsigned long long)(trunc_long(order >> 32, 32)) << 32);
+    /*
+    cerr << "order   = " << order << "\n";
+    cerr << "orderll = " << orderll << "\n";
+    */
+
+    unsigned int *rho, rho_prime, *r, *r_inv;
+    unsigned int l_y[SIZE_Y], l_g[SIZE_G];
+
+//cout << "rho = " << zz_rho << "\n";
+    ZZ zz_r = (to_ZZ(1) << (WORDS * BITS_PER_WORD));
+    ZZ_p::init(zz_r);
+    ZZ_p zz_p_rho_prime = -to_ZZ_p(1) / to_ZZ_p(modulus);
+//cout << "rho_prime = " << zz_p_rho_prime << "\n";
+    ZZ_p::init(modulus);
+    ZZ_p zz_p_r, zz_p_r_inv;
+    zz_p_r = to_ZZ_p(zz_r);
+    zz_p_r_inv = to_ZZ_p(1) / zz_p_r;
+
+    CBData cbdata(base, target, order);
+//cout << "r = " << zz_r << "\n";
+//cout << "r mod rho = " << rep(zz_p_r) << "\n";
+//cout << "r_inv = " << rep(zz_p_r_inv) << "\n";
+//cout << "r * r_inv = " << zz_p_r * zz_p_r_inv << "\n";
+//cout << "r * r_inv - rho * rho_prime = " << zz_r * rep(zz_p_r_inv) - zz_rho * rep(zz_p_rho_prime) << "\n";
+    rho = (unsigned int*) calloc((WORDS + 1), sizeof(unsigned int));
+    r = (unsigned int*) calloc((WORDS + 1), sizeof(unsigned int));
+    r_inv = (unsigned int*) calloc((WORDS + 1), sizeof(unsigned int));
+    BytesFromZZ((unsigned char *) rho, modulus, WORDS * sizeof(unsigned int));
+    BytesFromZZ((unsigned char *) &rho_prime, rep(zz_p_rho_prime), sizeof(unsigned int));
+    BytesFromZZ((unsigned char *) r,  rep(zz_p_r),  (WORDS + 1) * sizeof(unsigned int));
+    BytesFromZZ((unsigned char *) r_inv,  rep(zz_p_r_inv),  (WORDS + 1) * sizeof(unsigned int));
+    cudaMemcpyToSymbol(c_rho, rho, (WORDS + 1) * sizeof(unsigned int), 0, cudaMemcpyHostToDevice);
+    cudaMemcpyToSymbol(c_rho_prime, &rho_prime, sizeof(unsigned int), 0, cudaMemcpyHostToDevice);
+    cudaMemcpyToSymbol(c_r_inv, r_inv, (WORDS + 1) * sizeof(unsigned int), 0, cudaMemcpyHostToDevice);
+    checkCUDAError("memcpytosymbol");
+
+/*
+    if (nthreads % threadDimx) {
+	nthreads /= threadDimx;
+	nthreads *= threadDimx;
+	cerr << "Truncating to " << nthreads << " threads.\n";
+    }
+*/
+
+    if (nthreads % nblocks) {
+	cerr << "Error: " << nthreads << " not a multiple of " << nblocks << "\n";
+	exit(1);
+    }
+    const int threadsPerBlock = nthreads / nblocks;
+    GlobalThreadState *d_ts;
+    cudaMalloc((void **) &d_ts, nthreads * sizeof(GlobalThreadState));
+    GlobalThreadState *l_ts;
+    l_ts = (GlobalThreadState *) calloc(nthreads, sizeof(GlobalThreadState));
+    unsigned int *l_z;
+    l_z = (unsigned int *) calloc(nthreads, WORDS * sizeof(unsigned int));
+
+    //cudaPrintfInit((1<<20)*16);
+    DPstreamInit(1<<16);
+    struct timeval st, et;
+
+    memset(l_ts, 0, nthreads * sizeof(GlobalThreadState));
+    memset(l_z, 0, WORDS * sizeof(unsigned int));
+
+//cout << "y = " << rep(zz_p_y) << "\n";
+    ZZ_p zz_p_yr = target * zz_p_r;
+//cout << "y = " << rep(zz_p_y) << "\n";
+    ZZ_p zz_p_gr = base * zz_p_r;
+    BytesFromZZ((unsigned char *) l_y, rep(zz_p_yr), WORDS * sizeof(unsigned int));
+    BytesFromZZ((unsigned char *) l_g, rep(zz_p_gr), WORDS * sizeof(unsigned int));
+    cudaMemcpyToSymbol(c_y, l_y, (WORDS) * sizeof(unsigned int), 0, cudaMemcpyHostToDevice);
+    cudaMemcpyToSymbol(c_g, l_g, (WORDS) * sizeof(unsigned int), 0, cudaMemcpyHostToDevice);
+    //dump("y", l_y, WORDS);
+
+#ifdef CHECK_RESULTS
+    int totmult = 0;
+    bool fail = false;
+    vec_ZZ aexp, bexp;  // The expected values of a and b at completion
+    aexp.SetLength(nthreads);
+    bexp.SetLength(nthreads);
+#endif
+    ZZ a = RandomBits_ZZ(orderbits-1);
+    ZZ b = RandomBits_ZZ(orderbits-1);
+    ZZ astep = RandomBits_ZZ(orderbits-1);
+    ZZ bstep = RandomBits_ZZ(orderbits-1);
+    ZZ_p zz_p_x = power(base, a) * power(target, b);
+    ZZ_p zz_p_step = power(base, astep) * power(target, bstep);
+    ZZ_p zz_p_xr = zz_p_x * zz_p_r;
+    for (int t=0; t<nthreads; ++t)
+    {
+	zz_p_xr *= zz_p_step;
+	a += astep;
+	a %= order;
+	b += bstep;
+	b %= order;
+	BytesFromZZ((unsigned char *) l_ts[t].x, rep(zz_p_xr), WORDS * sizeof(unsigned int));
+	// dump("l_x", l_ts[t].x, WORDS);
+	BytesFromZZ((unsigned char *) l_ts[t].a, a, 2 * sizeof(unsigned int));
+	BytesFromZZ((unsigned char *) l_ts[t].b, b, 2 * sizeof(unsigned int));
+	//dump("x", l_ts[t].x, WORDS);
+//cout << "x = " << rep(zz_p_x) << "\n";
+//cout << "xr = " << rep(zz_p_xr) << "\n";
+	//dump("rho", rho, WORDS);
+#ifdef CHECK_RESULTS
+	for (int l = 0; l < nmult * nlaunch; ++l)
+	{
+	    unsigned int w = trunc_long(rep(zz_p_xr), 32);
+	    if (w < TWO_32_DIV_3) {
+		zz_p_xr *= base;
+		a++;
+	    } else if (w < TWO_32_DIV_3_X2) {
+		zz_p_xr *= target;
+		b++;
+	    } else {
+		zz_p_xr *= zz_p_xr * zz_p_r_inv;
+		a += a;
+		b += b;
+	    }
+	    if (a > order) a -= order;
+	    if (b > order) b -= order;
+	    ++totmult;
+	}     
+	BytesFromZZ((unsigned char *) (l_z + t * WORDS), rep(zz_p_xr), WORDS * sizeof(unsigned int));
+	aexp[t] = a;
+	bexp[t] = b;
+#endif
+//cout << "zr = " << rep(zz_p_xr) << "\n";
+	//dump("z", l_z + t * WORDS, WORDS);
+    }
+
+    cudaMemcpy(d_ts, l_ts, nthreads * sizeof(GlobalThreadState), cudaMemcpyHostToDevice);
+    dim3 tpb(threadDimx, threadsPerBlock/threadDimx);
+    gettimeofday(&st, NULL);
+    int launchcount = 0;
+#ifdef CHECK_RESULTS
+    for (int ln=0; ln<nlaunch; ++ln)
+#else
+    while(cbdata.found_collision == false)
+#endif
+    {
+	cerr << getpid() << " Launch " << ++launchcount << "...\n";
+	cudaMulmod<<< nblocks, nthreads/nblocks /*tpb*/, 0 >>>(d_ts, orderll);
+
+	cudaThreadSynchronize();
+	checkCUDAError("kernel launch");
+	DPstreamParse(&cbdata);
+	//cudaPrintfExtractDPE(dpcallback, &cbdata);
+	//cudaPrintfDisplay(stdout, true);
+	cerr << getpid() << "\n";
+    }
+    gettimeofday(&et, NULL);
+
+    totmicros = (unsigned long long)(et.tv_sec - st.tv_sec) * 1000000 + (et.tv_usec - st.tv_usec);
+#ifdef CHECK_RESULTS
+    cudaMemcpy(l_ts, d_ts, nthreads * sizeof(GlobalThreadState), cudaMemcpyDeviceToHost);
+    checkCUDAError("memcpy");
+
+    int j = 0;
+    for (int t = 0; t < nthreads; ++t)
+    {
+	unsigned int *l_Z = l_ts[t].x;
+	for (int i = 0; i < WORDS; i++)
+	{
+	    if (l_Z[i] != l_z[i + t * WORDS])
+	    {
+		fail = true;
+cout << i << ": " << l_Z[i] << " != " << l_z[i + t * WORDS] << "\n";
+	    }
+	}
+	if (fail)
+	{
+	    dump("d_z", l_Z, WORDS);
+	    dump("l_z", l_z + t * WORDS, WORDS);
+	}
+	ZZ ares = (to_ZZ(l_ts[t].a[1]) << 32) + to_ZZ(l_ts[t].a[0]);
+	ZZ bres = (to_ZZ(l_ts[t].b[1]) << 32) + to_ZZ(l_ts[t].b[0]);
+	if (ares != aexp[t] || bres != bexp[t]) {
+	    cerr << "ares = " << ares << "\n";
+	    cerr << "aexp = " << aexp[t] << "\n";
+	    cerr << "bres = " << bres << "\n";
+	    cerr << "bexp = " << bexp[t] << "\n";
+	}
+	++j;
+    }
+    if (!fail) {
+	cerr << "Results correct.\n";
+    }
+#endif
+
+    unsigned long long totnanos = totmicros * 1000;
+    cout << totmicros << " us / " << nthreads << " = " << totmicros / nthreads << " us / " << (nmult*launchcount) << " = " << totnanos / ((unsigned long long)nthreads * nmult * launchcount) << " ns\n";
+    //cudaPrintfEnd();
+    DPstreamEnd();
+
+    cudaFree(d_ts);
+    free(rho);
+    free(r_inv);
+    free(r);
+    free(l_ts);
+    free(l_z);
+
+    pbak.restore();
+
+    return cbdata.expon;
+}
+
+#ifdef TEST_CUDA
+int main(int argc, char** argv)
+{
+#ifdef DERANDOMIZE
+    SetSeed(to_ZZ(1));
+#else
+    // Initialize the prng with some randomness from the kernel
+    unsigned char randbuf[1024];
+    ifstream urand("/dev/urandom");
+    urand.read((char *) randbuf, sizeof(randbuf));
+    urand.close();
+    ZZ randzz = ZZFromBytes(randbuf, sizeof(randbuf));
+    SetSeed(randzz);
+#endif
+
+    ZZ rho, p, q;
+    vec_ZZ pfvec, qfvec;
+    cin >> rho >> p >> pfvec >> q >> qfvec;
+
+    if (argc > 1) nthreads = atoi(argv[1]);
+    if (argc > 2) nblocks = atoi(argv[2]);
+
+    ZZ_p::init(p);
+
+    for (int iter = 0; iter < 1; ++iter) {
+	// Create the subproblem for the f'th factor of p-1
+	int f = iter;
+
+	ZZ remorder = (p-1)/pfvec[f];
+	ZZ_p g = power(to_ZZ_p(2), remorder);
+	ZZ_p y = power(random_ZZ_p(), remorder);
+
+	// g should now be of order pfvec[0].  Check that.
+	if (g == 1 || power(g, pfvec[f]) != 1) {
+	    cerr << "base has the wrong order!\n";
+	    exit(1);
+	}
+
+	// Try to find the DL_g of y
+	cerr << "DL_" << g << "(" << y << ") mod " << p << "\n";
+	ZZ e = cuda_dl(g, y, pfvec[f], p);
+	cerr << "e = " << e << "\n";
+	cerr << ( (power(g,e) == y) ? "CORRECT!" : "INCORRECT!" ) << "\n";
+    }
+
+    return 0;
+}
+#endif